from numpy.testing import assert_array_almost_equal
from pytest import raises

from wonambi import Dataset

from .paths import ktlx_file


def test_xltek_data():
    d = Dataset(ktlx_file)
    data = d.read_data(begsam=1000, endsam=1001, chan=('FZ', ))
    assert_array_almost_equal(data.data[0][0, 0], -90.119315)


def test_xltek_marker():
    d = Dataset(ktlx_file)
    markers = d.read_markers()
    assert markers[0]['name'] == 'Start Recording (varkha)'
    assert int(markers[-1]['end']) == 1314


def test_xltek_videos():
    d = Dataset(ktlx_file)

    with raises(IndexError):
        d.read_videos(0, 2)

    videos, v_beg, v_end = d.read_videos(100, 200)
    assert len(videos) == 2
    assert v_beg == 58.410209
    assert v_end == 37.177808



"""Module to read open-ephys .continuous files

TODO
----
It assumes that the lenght of a block (i.e. record) is 1024 data points but this
might change in the future.
"""
from datetime import datetime
from logging import getLogger
import locale
from struct import unpack, calcsize
from math import ceil
from re import search, match
from xml.etree import ElementTree

from numpy import array, empty, nan, fromfile, unique, where, arange, hstack, vstack, concatenate

lg = getLogger(__name__)


HDR_LENGTH = 1024  # this is fixed
BLK_LENGTH = 1024  # this is currently fixed but might change in the future
BEG_BLK = '<qHH'  # time information and indices
DAT_FMT = '>' + BLK_LENGTH * 'h'  # actual data, big-endian
END_BLK = 10 * 'B'  # control values

BEG_BLK_SIZE = calcsize(BEG_BLK)
DAT_FMT_SIZE = calcsize(DAT_FMT)
BLK_SIZE = BEG_BLK_SIZE + DAT_FMT_SIZE + calcsize(END_BLK)

EVENT_TYPES = {
    3: 'TTL Event',
    5: 'Network Event',
    }
IGNORE_EVENTS = [
    'Network Event',
    ]


class OpenEphys:
    """Provide class OpenEphys, which can be used to read the folder
    containing Open-Ephys .continuous files

    Parameters
    ----------
    filename : Path
        Full path for the OpenEphys folder
    session : int
        session number (starting at 1, based on open-ephys convention)

    Attributes
    ----------
    channels : list of dict
        list of filenames referring to channels which are actually on disk
    blocks : 1D array
        length of each block, in samples (currently 1024)
    gain : 1D array
        gain to convert digital to microvolts (one value per channel)
    """
    def __init__(self, filename, session=1):

        if session == 1:
            self.session = ''
        else:
            self.session = f'_{session:d}'

        self.filename = filename.resolve()
        self.openephys_file = filename / f'Continuous_Data{self.session}.openephys'
        self.settings_xml = filename / f'settings{self.session}.xml'
        self.messages_file = filename / f'messages{self.session}.events'
        self.events_file = filename / f'all_channels{self.session}.events'

    def return_hdr(self):
        """Return the header for further use.

        Returns
        -------
        subj_id : str
            subject identification code
        start_time : datetime
            start time of the dataset
        s_freq : float
            sampling frequency
        chan_name : list of str
            list of all the channels
        n_samples : int
            number of samples in the dataset
        orig : dict
            currently empty for open-ephys
        """
        subj_id = self.filename.stem  # use directory name as subject name

        start_time = _read_date(self.settings_xml)

        self.recordings = _read_openephys(self.openephys_file)
        self.s_freq = self.recordings[0]['s_freq']
        channels = self.recordings[0]['channels']
        self.segments, self.messages, self.offset = _read_messages_events(self.messages_file)

        # only use channels that are actually in the folder
        chan_name = []
        self.channels = []
        gain = []
        for chan in channels:
            channel_filename = (self.filename / chan['filename'])

            if channel_filename.exists():
                chan_name.append(chan['name'])
                self.channels.append(channel_filename)

                ch_gain = _check_header(channel_filename, self.s_freq, self.offset)
                gain.append(ch_gain)

            else:
                lg.warning(f'could not find {chan["filename"]} in {self.filename}')
        self.gain = array(gain)

        # read data structure (recordings can have multiple segments)
        data_offset = [int(x['channels'][0]['position']) for x in self.recordings]
        for i in range(len(self.segments) - 1):
            self.segments[i]['length'] = int((data_offset[i + 1] - data_offset[i]) / BLK_SIZE * BLK_LENGTH)
            self.segments[i]['data_offset'] = data_offset[i]

        self.segments[-1]['data_offset'] = data_offset[-1]

        file_length = channel_filename.stat().st_size
        self.segments[-1]['length'] = int((file_length - data_offset[-1]) / BLK_SIZE * BLK_LENGTH)

        for seg in self.segments:
            seg['end'] = seg['start'] + seg['length']

        n_samples = self.segments[-1]['end']

        self.blocks_dat, self.blocks_offset = _prepare_blocks(self.segments)

        orig = {}

        return subj_id, start_time, self.s_freq, chan_name, n_samples, orig

    def return_dat(self, chan, begsam, endsam):
        """Read the data for some/all of the channels

        Parameters
        ----------
        chan : list of int
            indices of the channels to read
        begsam : int
            start sample to read
        endsam : int
            end sample to read (exclusive)

        Returns
        -------
        2D array
            chan X samples recordings
        """
        data_length = endsam - begsam
        dat = empty((len(chan), data_length))
        dat.fill(nan)

        all_blocks = _select_blocks(self.blocks_dat, begsam, endsam)
        for i_chan, sel_chan in enumerate(chan):
            with self.channels[sel_chan].open('rb') as f:
                for i_block in all_blocks:
                    i_dat = self.blocks_dat[i_block, :] - begsam
                    i_disk = self.blocks_offset[i_block].item()

                    # read only data (no timestamp or record marker)
                    f.seek(i_disk)
                    x = array(unpack(DAT_FMT, f.read(DAT_FMT_SIZE)))
                    beg_dat = max(i_dat[0], 0)
                    end_dat = min(i_dat[1], data_length)
                    beg_x = max(0, - i_dat[0])
                    segment_length = min(i_dat[1], data_length) - i_dat[0]
                    end_x = min(len(x), segment_length)
                    dat[i_chan, beg_dat:end_dat] = x[beg_x:end_x]

        return dat * self.gain[chan, None]

    def return_markers(self):
        """Read the markers from the .events file

        """
        all_markers = (
            self.messages
            + _segments_to_markers(self.segments, self.offset)
            + _read_all_channels_events(self.events_file, self.s_freq, self.offset)
            )
        return sorted(all_markers, key=lambda x: x['start'])


def _read_openephys(openephys_file):
    """Read the channel labels and their respective files from the
    'Continuous_Data.openephys' file

    Parameters
    ----------
    openephys_file : Path
        path to Continuous_Data.openephys inside the open-ephys folder

    Returns
    -------
    int
        sampling frequency
    list of dict
        list of channels containing the label, the filename and the gain

    TODO
    ----
    Check that all the recordings have the same channels and frequency
    """
    root = ElementTree.parse(openephys_file).getroot()

    recordings = []
    for recording in root:
        recordings.append({
            's_freq': float(recording.attrib['samplerate']),
            'number': int(recording.attrib['number']),
            'channels': [],
        })
        channels = []
        for processor in recording:
            for channel in processor:
                channels.append(channel.attrib)
        recordings[-1]['channels'] = channels

    return recordings


def _read_date(settings_file):
    """Get the data from the settings.xml file

    Parameters
    ----------
    settings_file : Path
        path to settings.xml inside open-ephys folder

    Returns
    -------
    datetime
        start time of the recordings

    Notes
    -----
    The start time is present in the header of each file. This might be useful
    if 'settings.xml' is not present.
    """
    locale.setlocale(locale.LC_TIME, 'en_US.utf-8')

    root = ElementTree.parse(settings_file).getroot()
    for e0 in root:
        if e0.tag == 'INFO':
            for e1 in e0:
                if e1.tag == 'DATE':
                    break

    return datetime.strptime(e1.text, '%d %b %Y %H:%M:%S')


def _read_header(filename):
    """Read the text header for each file

    Parameters
    ----------
    channel_file : Path
        path to single filename with the header

    Returns
    -------
    dict
        header
    int
        the timestamp of the first sample in the data. It's like an offset for
        the data. It's necessary to align with the clock time and the markers.
    """
    with filename.open('rb') as f:
        h = f.read(HDR_LENGTH).decode()

        header = {}
        for line in h.split('\n'):
            if '=' in line:
                key, value = line.split(' = ')
                key = key.strip()[7:]
                value = value.strip()[:-1]
                header[key] = value

        first_timestamp = unpack('q', f.read(8))[0]

    return header, first_timestamp


def _check_header(channel_file, s_freq, offset):
    """For each file, make sure that the header is consistent with the
    information in the text file.

    Parameters
    ----------
    channel_file : Path
        path to single filename with the header
    s_freq : int
        sampling frequency
    offset : int
        offset of the first timestamp

    Returns
    -------
    int
        gain from digital to microvolts (the same information is stored in
        the Continuous_Data.openephys but I trust the header for each file more.
    int
        the timestamp of the first sample in the data. It's like an offset for
        the data. It's necessary to align with the clock time and the markers.
    """
    hdr, first_timestamp = _read_header(channel_file)

    assert int(hdr['header_bytes']) == HDR_LENGTH
    assert int(hdr['blockLength']) == BLK_LENGTH
    assert int(hdr['sampleRate']) == s_freq
    assert first_timestamp == offset

    return float(hdr['bitVolts'])


def _segments_to_markers(segments, first_timestamp):
    mrk = []
    for i, seg in enumerate(segments):
        mrk.append({
            'name': f'START RECORDING #{i}',
            'chan': None,
            'start': seg['start'] / seg['s_freq'],
            'end': seg['start'] / seg['s_freq'],
        })
        mrk.append({
            'name': f'END RECORDING #{i}',
            'chan': None,
            'start': seg['end'] / seg['s_freq'],
            'end': seg['end'] / seg['s_freq'],
        })

    return mrk

def _read_messages_events(messages_file):
    messages = []
    segments = []
    header = True

    with messages_file.open() as f:
        for l in f:

            m_time = search(r'\d+ Software time: (\d+)@\d+Hz', l)
            m_start = search(r'start time: (\d+)@(\d+)Hz', l)
            m_event = match(r'(\d+) (.+)', l)

            if m_time:
                # ignore Software time
                pass

            elif m_start:
                if header:
                    offset = int(m_start.group(1))
                    s_freq = int(m_start.group(2))
                    header = False

                segments.append({
                    'start': int(m_start.group(1)) - offset,
                    's_freq': int(m_start.group(2)),
                    })

            elif m_event:
                time = int(m_event.group(1))
                messages.append({
                    'name': m_event.group(2),
                    'start': (time - offset) / s_freq,
                    'end': (time - offset) / s_freq,
                    'chan': None,
                })

    return segments, messages, offset


def _read_all_channels_events(events_file, s_freq, offset):

    file_read = [
        ('timestamps', '<i8'),
        ('sampleNum', '<i2'),
        ('eventType', '<u1'),
        ('nodeId', '<u1'),
        ('eventId', '<u1'),
        ('channel', '<u1'),
        ('recordingNumber', '<u2'),
        ]

    with events_file.open('rb') as f:
        f.seek(HDR_LENGTH)
        evt = fromfile(f, file_read)

    mrk = []
    for evt_type in unique(evt['eventType']):
        timestamps = evt[evt['eventType'] == evt_type]['timestamps']
        onsets = timestamps[::2]
        offsets = timestamps[1::2]

        for i_on, i_off in zip(onsets, offsets):
            mrk.append({
                'name': EVENT_TYPES[evt_type],
                'start': (i_on - offset) / s_freq,
                'end': (i_off - offset) / s_freq,
                'chan': None,
            })

    # skip some events (like Network Events) which are not very useful
    mrk = [evt for evt in mrk if not evt['name'] in IGNORE_EVENTS]

    return mrk


def _prepare_blocks(segments):

    blocks_dat = []
    blocks_offset = []

    for seg in segments:
        n_blocks = ceil(seg['length'] / BLK_LENGTH)

        blocks_dat.append(vstack((
            arange(n_blocks) * BLK_LENGTH + seg['start'],
            arange(n_blocks) * BLK_LENGTH + BLK_LENGTH + seg['start'],
            )))
        blocks_offset.append(
            arange(n_blocks) * BLK_SIZE + seg['data_offset'] + BEG_BLK_SIZE)

    blocks_dat = hstack(blocks_dat).T
    blocks_offset = concatenate(blocks_offset)

    return blocks_dat, blocks_offset


def _select_blocks(blocks_dat, begsam, endsam):
    all_blocks = ((blocks_dat[:, 1] - begsam) > 0) & ((endsam - blocks_dat[:, 0]) > 0)
    return where(all_blocks)[0]



from datetime import datetime
from logging import getLogger
from os import SEEK_CUR
from pathlib import Path
from struct import unpack
from xml.etree.ElementTree import parse

from numpy import (append, asarray, cumsum, diff, empty, nan, sum,
                   where, ndarray, unique)

from .utils import DEFAULT_DATETIME


lg = getLogger(__name__)


class EgiMff:
    """Basic class to read the data.

    Parameters
    ----------
    filename : path to file
        the name of the filename or directory

    """
    def __init__(self, filename):
        self.filename = Path(filename)
        self._signal = []
        self._block_hdr = []
        self._i_data = []
        self._nchan_signal1 = []  # n of channels in signal1
        self._n_samples = []
        self._orig = {}

    def return_hdr(self):
        """Return the header for further use.

        Returns
        -------
        subj_id : str
            subject identification code
        start_time : datetime
            start time of the dataset
        s_freq : float
            sampling frequency
        chan_name : list of str
            list of all the channels
        n_samples : int
            number of samples in the dataset
        orig : dict
            additional information taken directly from the header

        """
        orig = {}
        for xml_file in self.filename.glob('*.xml'):
            if xml_file.stem[0] != '.':
                orig[xml_file.stem] = parse_xml(str(xml_file))

        signals = sorted(self.filename.glob('signal*.bin'))

        for signal in signals:
            block_hdr, i_data = read_all_block_hdr(signal)
            self._signal.append(signal)
            self._block_hdr.append(block_hdr)
            self._i_data.append(i_data)
            n_samples = asarray([x['n_samples'][0] for x in block_hdr], 'q')
            self._n_samples.append(n_samples)

        try:
            subj_id = orig['subject'][0][0]['name']
        except KeyError:
            subj_id = ''
        try:
            start_time = datetime.strptime(orig['info'][0]['recordTime'][:26],
                                           '%Y-%m-%dT%H:%M:%S.%f')
        except KeyError:
            start_time = DEFAULT_DATETIME
        self.start_time = start_time

        videos = (list(self.filename.glob('*.mp4')) +  # as described in specs
                  list(self.filename.glob('*.mov')))  # actual example
        videos = [x for x in videos if x.stem[0] != '.']  # remove hidden files

        if len(videos) > 1:
            lg.warning('More than one video present: ' + ', '.join(videos))
        self._videos = videos

        # it only works if they have all the same sampling frequency
        s_freq = [x[0]['freq'][0] for x in self._block_hdr]
        assert all([x == s_freq[0] for x in s_freq])
        SIGNAL = 0
        s_freq = self._block_hdr[SIGNAL][0]['freq'][0]
        n_samples = sum(self._n_samples[SIGNAL])

        chan_name, self._nchan_signal1 = _read_chan_name(orig)
        self._orig = orig

        return subj_id, start_time, s_freq, chan_name, n_samples, orig

    def return_dat(self, chan, begsam, endsam):
        """Return the data as 2D numpy.ndarray.

        Parameters
        ----------
        chan : list
            indices of the channels to read
        begsam : int
            index of the first sample
        endsam : int
            index of the last sample

        Returns
        -------
        numpy.ndarray
            A 2d matrix, with dimension chan X samples

        Notes
        -----
        This format is tricky for both channels and samples. For the samples,
        we just use the boundaries in the block header. For the channels, we
        assume that there are max two signals, one EEG and one PIB box. We
        just use the boundary between them to define if a channel belongs to
        the first group or to the second.

        TODO
        ----
        use wonambi.ioeeg.utils._select_blocks here, but you need to test it
        with a PIB box.
        """
        assert begsam < endsam

        data = empty((len(chan), endsam - begsam))
        data.fill(nan)

        chan = asarray(chan)

        # we assume there are only two signals
        signals_to_read = []
        if (chan < self._nchan_signal1).any():
            signals_to_read.append(0)
        if (chan >= self._nchan_signal1).any():
            signals_to_read.append(1)

        for one_signal in signals_to_read:
            if one_signal == 0:
                i_chan_data = chan < self._nchan_signal1
                i_chan_rec = chan[i_chan_data]

            if one_signal == 1:
                i_chan_data = chan >= self._nchan_signal1
                i_chan_rec = chan[i_chan_data] - self._nchan_signal1

            x = self._n_samples[one_signal]
            x1 = cumsum(append(0, x))

            # begrec is -1 when begsam is before start of the recordings
            begrec = where(begsam < x1)[0][0] - 1
            try:
                endrec = where(endsam < x1)[0][0] - 1
            except IndexError:
                endrec = len(x)

            f = self._signal[one_signal].open('rb')

            i0 = 0
            for rec in range(begrec, endrec + 1):

                # if begsam is before start of the recordings, we just shift the baseline
                if rec == -1:
                    i0 = - begsam
                    continue

                # if endsam is after end of the recordings, we just stop here
                if rec == len(self._n_samples[one_signal]):
                    break

                if rec == begrec:
                    begpos_rec = begsam - x1[rec]
                else:
                    begpos_rec = 0

                if rec == endrec:
                    endpos_rec = endsam - x1[rec]
                else:
                    endpos_rec = x[rec]

                i1 = i0 + endpos_rec - begpos_rec

                lg.debug('data {: 8d}-{: 8d}, rec ({}) {: 5d} - {: 5d}'.format(i0, i1, rec, begpos_rec, endpos_rec))

                rec_dat = _read_block(f,
                                      self._block_hdr[one_signal][rec],
                                      self._i_data[one_signal][rec])

                data[i_chan_data, i0:i1] = rec_dat[i_chan_rec,
                                                   begpos_rec:endpos_rec]
                i0 = i1

            f.close()

        return data

    def return_markers(self):
        """"""
        xml_files = self._orig.keys()
        xml_events = [x for x in xml_files if x[:7] == 'Events_']

        markers = []
        for xml in xml_events:
            for event in self._orig[xml][2:]:
                event_start = datetime.strptime(event['beginTime'][:26],
                                                '%Y-%m-%dT%H:%M:%S.%f')
                start = (event_start - self.start_time).total_seconds()

                marker = {'name': event['code'],
                          'start': start,
                          'end': start + float(event['duration']) / 1e9,
                          'chan': None,
                          }
                markers.append(marker)

        return markers

    def return_videos(self, begtime, endtime):
        """It returns the videos and beginning and end time of the video
        segment. The MFF video format is not well documented. As far as I can
        see, the manual 20150805 says that there might be multiple .mp4 files
        but I only see one .mov file (and no way to specify which video file to
        read). In addition, there is a file "po_videoSyncups.xml" which seems
        to contain some time information, but the sampleTime does not start at
        zero, but at a large number. I don't know how to use the info in
        po_videoSyncups.xml.

        Parameters
        ----------
        begtime : float
            start time of the period of interest
        endtime : float
            end time of the period of interest

        Returns
        -------
        list of one path
            list with only one element
        float
            start time of the video
        float
            end time of the video
        """
        try:
            self._orig['po_videoSyncups']
        except KeyError:
            raise OSError('No po_videoSyncups.xml in folder to sync videos')

        if not self._videos:
            raise OSError('No mp4 video files')

        mp4_file = self._videos[:1]  # make clear we only use the first video

        return mp4_file, begtime, endtime


def _read_block(filename, block_hdr, i):
    f = filename

    # Can we assume constant depth across blocks?
    depth = unique(block_hdr['depth'])
    assert(len(depth) == 1)
    n_bytes = (depth[0] / 8).astype('B')

    if n_bytes == 2:
        data_type = '<h'
    elif n_bytes == 4:
        data_type = '<f'
    elif n_bytes == 8:
        data_type = '<d'
    else:
        raise ValueError("Invalid depth parameter.")

    f.seek(i)

    return ndarray((block_hdr['n_signals'], block_hdr['n_samples'][0]),
                   data_type,
                   f.read(sum(n_bytes * block_hdr['n_samples'])))


def read_block_hdr(f):

    version = unpack('<I', f.read(4))[0]
    hdr_size = unpack('<I', f.read(4))[0]
    data_size = unpack('<I', f.read(4))[0]
    n_signals = unpack('<I', f.read(4))[0]

    offset = asarray(unpack('<' + 'I' * n_signals, f.read(4 * n_signals)), 'I')

    depth = empty(n_signals, 'B')
    freq = empty(n_signals, 'I')
    for j in range(n_signals):
        depth[j] = unpack('<B', f.read(1))[0]
        freq[j] = unpack('<I', f.read(3) + b'\x00')[0]

    n_samples = diff(offset)
    n_samples = append(n_samples, data_size - offset[-1])
    n_samples = (n_samples / (depth / 8)).astype('I')

    opt_hdr_size = unpack('<I', f.read(4))[0]
    if opt_hdr_size > 0:
        opt_hdr_type = unpack('<I', f.read(4))[0]
        n_blocks = unpack('<Q', f.read(8))[0]
        n_smp = unpack('<Q', f.read(8))[0]  # unreliable
        n_signals_opt = unpack('<I', f.read(4))[0]
    else:
        n_blocks = None
        n_smp = None
        n_signals_opt = None

    opt_hdr = {'n_blocks': n_blocks,
               'n_smp': n_smp,
               'n_signals_opt': n_signals_opt}

    hdr = {'version': version,
           'hdr_size': hdr_size,
           'data_size': data_size,
           'n_signals': n_signals,
           'offset': offset,
           'depth': depth,
           'freq': freq,
           'n_samples': n_samples,
           'opt_hdr_size': opt_hdr_size,
           'opt_hdr': opt_hdr,
           }

    return hdr


def read_all_block_hdr(filename):

    with filename.open('rb') as f:

        block_hdr = []
        i_data = []

        while True:
            version_bytes = f.read(4)
            if not version_bytes:
                break
            version = unpack('<I', version_bytes)[0]

            if version:
                f.seek(-4, SEEK_CUR)  # re-read version
                block_hdr.append(read_block_hdr(f))

            else:
                block_hdr.append(block_hdr[-1])

            i_data.append(f.tell())
            f.seek(block_hdr[-1]['data_size'], SEEK_CUR)

        i_data = asarray(i_data, 'q')

    return block_hdr, i_data


def parse_xml(xml_file):
    xml = parse(xml_file)
    root = xml.getroot()
    return xml2list(root)


def ns(s):
    """remove namespace, but only it there is a namespace to begin with"""
    if '}' in s:
        return '}'.join(s.split('}')[1:])
    else:
        return s


def xml2list(root):
    output = []

    for element in root:

        if element:

            if element[0].tag != element[-1].tag:
                output.append(xml2dict(element))
            else:
                output.append(xml2list(element))

        elif element.text:
            text = element.text.strip()
            if text:
                tag = ns(element.tag)
                output.append({tag: text})

    return output


def xml2dict(root):
    """Use functions instead of Class and remove namespace based on:
    http://stackoverflow.com/questions/2148119
    """
    output = {}
    if root.items():
        output.update(dict(root.items()))

    for element in root:
        if element:
            if len(element) == 1 or element[0].tag != element[1].tag:
                one_dict = xml2dict(element)
            else:
                one_dict = {ns(element[0].tag): xml2list(element)}

            if element.items():
                one_dict.update(dict(element.items()))
            output.update({ns(element.tag): one_dict})

        elif element.items():
            output.update({ns(element.tag): dict(element.items())})

        else:
            output.update({ns(element.tag): element.text})

    return output


def _read_chan_name(orig):
    """Read channel labels, which can be across xml files.

    Parameters
    ----------
    orig : dict
        contains the converted xml information

    Returns
    -------
    list of str
        list of channel names
    ndarray
        vector to indicate to which signal a channel belongs to

    Notes
    -----
    This assumes that the PIB Box is the second signal.
    """
    sensors = orig['sensorLayout'][1]

    eeg_chan = []
    for one_sensor in sensors:
        if one_sensor['type'] in ('0', '1'):
            if one_sensor['name'] is not None:
                eeg_chan.append(one_sensor['name'])
            else:
                eeg_chan.append(one_sensor['number'])

    pns_chan = []
    if 'pnsSet' in orig:
        pnsSet = orig['pnsSet'][1]
        for one_sensor in pnsSet:
            pns_chan.append(one_sensor['name'])

    return eeg_chan + pns_chan, len(eeg_chan)



from wonambi import Wonambi
from PyQt5.QtWidgets import QAction
from os.path import basename, splitext
import csv
from pytest import approx

from .test_scroll_data import (channel_make_group,
                               find_in_qt,
                               screenshot,
                               )

from .paths import (annot_psg_path,
                    gui_file,
                    GUI_PATH,
                    analysis_export_path,
                    EXPORTED_PATH,
                    )

def test_widget_analysis_frequency(qtbot):

    w = Wonambi()
    qtbot.addWidget(w)

    w.info.open_dataset(str(gui_file))
    channel_make_group(w)
    w.channels.button_apply.click()
    w.notes.update_notes(annot_psg_path)

    w.notes.action['analyze']
    ad = w.analysis_dialog
    ad.update_evt_types()
    ad.update_groups()
    ad.grab().save(str(GUI_PATH / 'analysis_01_dialog.png'))
    ad.tab_evt.grab().save(str(GUI_PATH / 'analysis_03_event.png'))

    ad.filename = analysis_export_path
    ad.chunk['epoch'].setChecked(True)
    ad.lock_to_staging.set_value(True)
    ad.idx_chan.setCurrentRow(0)
    ad.idx_stage.setCurrentRow(2)
    ad.trans['diff'].set_value(True)
    freq = ad.frequency
    freq['export_full'].set_value(True)
    freq['export_band'].set_value(True)
    freq['band'].set_value('[[0.5-4],[4-8],[10-16],[30-50]]')
    freq['prep'].set_value(True)
    freq['norm'].set_value('by integral of each segment')
    ad.tab_freq.grab().save(str(GUI_PATH / 'analysis_02_freq.png'))
    assert ad.nseg == 127

    ad.button_clicked(ad.idx_ok)
    freq_path = EXPORTED_PATH / (splitext(basename(analysis_export_path))[0] +
                                 '_freq_full.csv')
    with open(freq_path) as f:
        reader = csv.reader(f)
        rows = [row for row in reader]
    assert approx(float(rows[2][44])) == 0.01862944649426
    assert approx(float(rows[4][16])) == -4.28072153999

    band_path = EXPORTED_PATH / (splitext(basename(analysis_export_path))[0] +
                                 '_freq_band.csv')
    with open(band_path) as f:
        reader = csv.reader(f)
        rows = [row for row in reader]
    assert approx(float(rows[2][10])) == 0.0944796084801
    assert approx(float(rows[12][11])) == 0.063348951380


def test_widget_analysis_fooof(qtbot):

    w = Wonambi()
    qtbot.addWidget(w)

    w.info.open_dataset(str(gui_file))
    channel_make_group(w)
    w.channels.button_apply.click()
    w.notes.update_notes(annot_psg_path)

    w.notes.action['analyze']
    ad = w.analysis_dialog
    ad.update_evt_types()
    ad.update_groups()

    ad.filename = analysis_export_path
    ad.chunk['segment'].setChecked(True)
    ad.idx_chan.setCurrentRow(0)
    ad.idx_stage.setCurrentRow(1)
    freq = ad.frequency
    freq['fooof_on'].set_value(True)
    freq['prep'].set_value(False)
    freq['step'].set_value(True)
    freq['step_val'].set_value(0.5)
    freq['norm'].set_value('none')
    assert ad.nseg == 33

    ad.button_clicked(ad.idx_ok)
    fooof_path = EXPORTED_PATH / (splitext(basename(analysis_export_path))[0] +
                                  '_freq_fooof.csv')
    with open(fooof_path) as f:
        reader = csv.reader(f)
        rows = [row for row in reader]
    assert approx(float(rows[11][1])) == 12.613759924714094
    assert approx(float(rows[11][2])) == 0.24356640365481338


def test_widget_analysis_event(qtbot):

    w = Wonambi()
    qtbot.addWidget(w)

    w.info.open_dataset(str(gui_file))
    channel_make_group(w)
    w.channels.button_apply.click()
    w.notes.update_notes(annot_psg_path)
    w.traces.go_to_epoch(test_text_str='23:34:45')

    w.notes.delete_eventtype(test_type_str='spindle')
    w.notes.new_eventtype(test_type_str='spindle')
    w.notes.action['new_event'].setChecked(True)
    w.notes.add_event('spindle', (24293.01, 24294.65), 'EEG Pz-Oz (scalp)')
    w.notes.add_event('spindle', (24288.01, 24288.90), 'EEG Fpz-Cz (scalp)')
    w.notes.add_event('spindle', (24290.5, 24291.00), 'EEG Fpz-Cz (scalp)')

    w.notes.action['analyze']
    ad = w.analysis_dialog
    ad.update_evt_types()
    ad.update_groups()

    ad.filename = analysis_export_path
    ad.chunk['event'].setChecked(True)
    ad.idx_evt_type.setCurrentRow(1)
    ad.idx_chan.setCurrentRow(0)
    evt = ad.event
    evt['global']['count'].set_value(True)
    evt['global']['density'].set_value(True)
    evt['f1'].set_value(10)
    evt['f2'].set_value(16)
    evt['global']['all_local'].set_value(True)
    evt['sw']['avg_slope'].set_value(True)
    evt['sw']['max_slope'].set_value(True)
    ad.check_all_local()

    ad.button_clicked(ad.idx_ok)
    w.notes.delete_eventtype(test_type_str='spindle')
    w.close()

    evt_path = EXPORTED_PATH / (splitext(basename(analysis_export_path))[0] +
                                 '_params.csv')
    with open(evt_path) as f:
        reader = csv.reader(f)
        rows = [row for row in reader]
    assert approx(float(rows[1][1])) == 2
    assert approx(float(rows[2][1])) == 0.0020768431983
    assert approx(float(rows[4][15])) == 9.81616222270557
    assert approx(float(rows[9][13])) ==  7.052493472354065


def test_widget_notes_export_csv(qtbot):

    w = Wonambi()
    qtbot.addWidget(w)

    w.info.open_dataset(str(gui_file))
    w.notes.update_notes(annot_psg_path)

    menubar = w.menuBar()

    act_annot = find_in_qt(menubar, QAction, 'Annotations')
    menubar.setActiveAction(act_annot)
    act_annot.menu().setActiveAction(w.notes.action['export_sleepstats'])

    screenshot(w, 'analysis_02_statistics.png')
    w.close()



"""Module to reject bad channels and bad epochs.
"""
from logging import getLogger

lg = getLogger(__name__)


def rejectbadchan():
    """Class to reject bad channels.
    """
    pass

def remove_artf_evts(times, annot, chan=None, name=None, min_dur=0.1):
    """Correct times to remove events marked 'Artefact'.

    Parameters
    ----------
    times : list of tuple of float
        the start and end times of each segment
    annot : instance of Annotations
        the annotation file containing events and epochs
    chan : str, optional
        full name of channel on which artefacts were marked. Channel format is 
        'chan_name (group_name)'. If None, artefacts from any channel will be
        removed.
    name : str or list of str, optional
        name of the event type(s) to be rejected. If None, defaults to 
        'Artefact'.
    min_dur : float
        resulting segments, after concatenation, are rejected if shorter than
        this duration

    Returns
    -------
    list of tuple of float
        the new start and end times of each segment, with artefact periods 
        taken out            
    """    
    new_times = times
    beg = times[0][0]
    end = times[-1][-1]
    chan = (chan, '') if chan else None # '' is for channel-global artefacts
    
    if name is not None:
        if isinstance(name, list):
            evt_type_list = name
        elif isinstance(name, str):
            evt_type_list = [name]
        else:
            raise TypeError(
                    "Argument 'name' must be str, list of str, or None.")
    else:
        evt_type_list = ['Artefact']            
    
    artefact = []
    for evt_type in evt_type_list:
        artefact.extend(annot.get_events(name=evt_type, time=(beg, end), 
                                         chan=chan))

    if artefact:
        artefact = sorted(artefact, key=lambda x: x['start'])
        new_times = []
        
        for seg in times:
            reject = False
            new_seg = True
            
            while new_seg is not False:
                if type(new_seg) is tuple:
                    seg = new_seg
                end = seg[1]
            
                for artf in artefact:
                    
                    if artf['start'] <= seg[0] and seg[1] <= artf['end']:
                        reject = True
                        new_seg = False
                        break
                    
                    a_starts_in_s = seg[0] <= artf['start'] < seg[1]
                    a_ends_in_s = seg[0] < artf['end'] <= seg[1]
                    
                    if a_ends_in_s and not a_starts_in_s:
                        seg = artf['end'], seg[1]
                        
                    elif a_starts_in_s:
                        seg = seg[0], artf['start']
    
                        if a_ends_in_s:
                            new_seg = artf['end'], end
                        else:
                            new_seg = False
                        break
                    
                    new_seg = False
            
                if reject is False and seg[1] - seg[0] >= min_dur:
                    new_times.append(seg)
        
    return new_times
